@@ -39,7 +39,11 @@
 #define PANGO_WIN32_IS_FONT_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), PANGO_TYPE_WIN32_FONT))
 #define PANGO_WIN32_FONT_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), PANGO_TYPE_WIN32_FONT, PangoWin32FontClass))
 
+#define PANGO_WIN32_UNKNOWN_FLAG 0x10000000
+
 HDC pango_win32_hdc;
+OSVERSIONINFO pango_win32_os_version_info;
+gboolean pango_win32_debug = FALSE;
 
 typedef struct _PangoWin32FontClass   PangoWin32FontClass;
 
@@ -48,10 +52,6 @@ struct _PangoWin32FontClass
   PangoFontClass parent_class;
 };
 
-static PangoFontClass *parent_class;	/* Parent class structure for PangoWin32Font */
-
-static void pango_win32_font_class_init (PangoWin32FontClass *class);
-static void pango_win32_font_init       (PangoWin32Font      *win32font);
 static void pango_win32_font_dispose    (GObject             *object);
 static void pango_win32_font_finalize   (GObject             *object);
 
@@ -59,7 +59,8 @@ static PangoFontDescription *pango_win32_font_describe          (PangoFont
 static PangoCoverage        *pango_win32_font_get_coverage      (PangoFont        *font,
 								 PangoLanguage    *lang);
 static void                  pango_win32_font_calc_coverage     (PangoFont        *font,
-								 PangoCoverage    *coverage);
+								 PangoCoverage    *coverage,
+								 PangoLanguage    *lang);
 static PangoEngineShape     *pango_win32_font_find_shaper       (PangoFont        *font,
 								 PangoLanguage    *lang,
 								 guint32           ch);
@@ -91,7 +92,10 @@ pango_win32_get_hfont (PangoFont *font)
       win32font->hfont = pango_win32_font_cache_load (cache, &win32font->logfont);
       if (!win32font->hfont)
 	{
-	  g_warning ("Cannot load font '%s\n", win32font->logfont.lfFaceName);
+	  gchar *face_utf8 = g_locale_to_utf8 (win32font->logfont.lfFaceName,
+					       -1, NULL, NULL, NULL);
+	  g_warning ("Cannot load font '%s\n", face_utf8);
+	  g_free (face_utf8);
 	  return NULL;
 	}
 
@@ -134,33 +138,7 @@ pango_win32_get_context (void)
   return result;
 }
 
-static GType
-pango_win32_font_get_type (void)
-{
-  static GType object_type = 0;
-
-  if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (PangoWin32FontClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) pango_win32_font_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoWin32Font),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) pango_win32_font_init,
-      };
-      
-      object_type = g_type_register_static (PANGO_TYPE_FONT,
-                                            "PangoWin32Font",
-                                            &object_info, 0);
-    }
-  
-  return object_type;
-}
+G_DEFINE_TYPE (PangoWin32Font, pango_win32_font, PANGO_TYPE_FONT)
 
 static void 
 pango_win32_font_init (PangoWin32Font *win32font)
@@ -170,14 +148,59 @@ pango_win32_font_init (PangoWin32Font *win32font)
   win32font->glyph_info = g_hash_table_new_full (NULL, NULL, NULL, g_free);
 }
 
+/**
+ * pango_win32_get_dc:
+ * 
+ * Obtains a handle to the Windows device context that is used by Pango.
+ * 
+ * Return value: A handle to the Windows device context that is used by Pango.
+ **/
+HDC
+pango_win32_get_dc (void)
+{
+  if (pango_win32_hdc == NULL)
+    {
+      pango_win32_hdc = CreateDC ("DISPLAY", NULL, NULL, NULL);
+      memset (&pango_win32_os_version_info, 0,
+	      sizeof (pango_win32_os_version_info));
+      pango_win32_os_version_info.dwOSVersionInfoSize =
+	sizeof (OSVERSIONINFO);
+      GetVersionEx (&pango_win32_os_version_info);
+
+      /* Also do some generic pangowin32 initialisations... this function
+       * is a suitable place for those as it is called from a couple
+       * of class_init functions.
+       */
+#ifdef PANGO_WIN32_DEBUGGING
+      if (getenv ("PANGO_WIN32_DEBUG") != NULL)
+	pango_win32_debug = TRUE;
+#endif      
+    }
+
+  return pango_win32_hdc;
+}  
+
+/**
+ * pango_win32_get_debug_flag:
+ *
+ * Returns wether debugging is turned on.
+ * 
+ * Returns: %TRUE if debugging is turned on.
+ * 
+ * Since: 1.2
+ */
+gboolean
+pango_win32_get_debug_flag (void)
+{
+  return pango_win32_debug;
+}
+
 static void
 pango_win32_font_class_init (PangoWin32FontClass *class)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (class);
   PangoFontClass *font_class = PANGO_FONT_CLASS (class);
 
-  parent_class = g_type_class_peek_parent (class);
-  
   object_class->finalize = pango_win32_font_finalize;
   object_class->dispose = pango_win32_font_dispose;
   
@@ -187,8 +210,7 @@ pango_win32_font_class_init (PangoWin32FontClass *class)
   font_class->get_glyph_extents = pango_win32_font_get_glyph_extents;
   font_class->get_metrics = pango_win32_font_get_metrics;
 
-  if (pango_win32_hdc == NULL)
-    pango_win32_hdc = CreateDC ("DISPLAY", NULL, NULL, NULL);
+  pango_win32_get_dc ();
 }
 
 PangoWin32Font *
@@ -204,7 +226,7 @@ pango_win32_font_new (PangoFontMap  *fontmap,
   result = (PangoWin32Font *)g_object_new (PANGO_TYPE_WIN32_FONT, NULL);
   
   result->fontmap = fontmap;
-  g_object_ref (G_OBJECT (fontmap));
+  g_object_ref (fontmap);
 
   result->size = size;
   pango_win32_make_matching_logfont (fontmap, lfp, size,
@@ -224,20 +246,40 @@ pango_win32_font_new (PangoFontMap  *fontmap,
  * Render a PangoGlyphString onto a Windows DC
  */
 void 
-pango_win32_render (HDC                hdc,
-		    PangoFont         *font,
-		    PangoGlyphString  *glyphs,
-		    int                x, 
-		    int                y)
+pango_win32_render (HDC               hdc,
+		    PangoFont        *font,
+		    PangoGlyphString *glyphs,
+		    int               x, 
+		    int               y)
 {
-  HFONT old_hfont = NULL;
-  HFONT hfont;
-  int i;
+  HFONT hfont, old_hfont = NULL;
+  int i, j, num_valid_glyphs;
   guint16 *glyph_indexes;
-  INT *dX;
+  gint *dX;
+  gint this_x;
+  PangoGlyphUnit start_x_offset, x_offset, next_x_offset, cur_y_offset;
 
   g_return_if_fail (glyphs != NULL);
 
+#ifdef PANGO_WIN32_DEBUGGING
+  if (pango_win32_debug)
+    {
+      PING (("num_glyphs:%d", glyphs->num_glyphs));
+      for (i = 0; i < glyphs->num_glyphs; i++)
+	{
+	  printf (" %d:%d", glyphs->glyphs[i].glyph, glyphs->glyphs[i].geometry.width);
+	  if (glyphs->glyphs[i].geometry.x_offset != 0 || 
+	      glyphs->glyphs[i].geometry.y_offset != 0)
+	    printf (":%d,%d", glyphs->glyphs[i].geometry.x_offset,
+		    glyphs->glyphs[i].geometry.y_offset);
+	}
+      printf ("\n");
+    }
+#endif
+
+  if (glyphs->num_glyphs == 0)
+    return;
+
   hfont = pango_win32_get_hfont (font);
   if (!hfont)
     return;
@@ -247,17 +289,107 @@ pango_win32_render (HDC                hdc,
   glyph_indexes = g_new (guint16, glyphs->num_glyphs);
   dX = g_new (INT, glyphs->num_glyphs);
 
-  for (i = 0; i <glyphs->num_glyphs; i++)
+  /* Render glyphs using one ExtTextOutW() call for each run of glyphs
+   * that have the same y offset. The big majoroty of glyphs will have
+   * y offset of zero, so in general, the whole glyph string will be
+   * rendered by one call to ExtTextOutW().
+   *
+   * In order to minimize buildup of rounding errors, we keep track of
+   * where the glyphs should be rendered in PangoGlyphUnits, and round
+   * to pixels separately for each glyph,
+   */
+
+  i = 0;
+
+  /* Outer loop through all glyphs in string */
+  while (i < glyphs->num_glyphs)
     {
-      glyph_indexes[i] = glyphs->glyphs[i].glyph;
-      dX[i] = glyphs->glyphs[i].geometry.width / PANGO_SCALE;
+      cur_y_offset = glyphs->glyphs[i].geometry.y_offset;
+      num_valid_glyphs = 0;
+      x_offset = 0;
+      start_x_offset = glyphs->glyphs[i].geometry.x_offset;
+      this_x = PANGO_PIXELS (start_x_offset);
+
+      /* Inner loop through glyphs with the same y offset, or code
+       * point zero (just spacing).
+       */
+      while (i < glyphs->num_glyphs &&
+	     (glyphs->glyphs[i].glyph == 0 || cur_y_offset == glyphs->glyphs[i].geometry.y_offset))
+	{
+	  if (glyphs->glyphs[i].glyph == 0)
+	    {
+	      /* Code point 0 glyphs should not be rendered, but their
+	       * indicated width (set up by PangoLayout) should be taken
+	       * into account.
+	       */
+
+	      /* If the string starts with spacing, must shift the
+	       * starting point for the glyphs actually rendered.  For
+	       * spacing in the middle of the glyph string, add to the dX
+	       * of the previous glyph to be rendered.
+	       */
+	      if (num_valid_glyphs == 0)
+		start_x_offset += glyphs->glyphs[i].geometry.width;
+	      else
+		{
+		  x_offset += glyphs->glyphs[i].geometry.width;
+		  dX[num_valid_glyphs-1] = PANGO_PIXELS (x_offset) - this_x;
+		}
+	    }
+	  else
+	    {
+	      if (glyphs->glyphs[i].glyph & PANGO_WIN32_UNKNOWN_FLAG)
+		{
+		  /* Glyph index is actually the char value that doesn't
+		   * have any glyph (ORed with the flag). We should really
+		   * do the same that pango_xft_real_render() does: render
+		   * a box with the char value in hex inside it in a tiny
+		   * font. Later. For now, use the TrueType invalid glyph
+		   * at 0.
+		   */
+		  glyph_indexes[num_valid_glyphs] = 0;
+		}
+	      else
+		glyph_indexes[num_valid_glyphs] = glyphs->glyphs[i].glyph;
+
+	      x_offset += glyphs->glyphs[i].geometry.width;
+
+	      /* If the next glyph has an X offset, take that into consideration now */
+	      if (i < glyphs->num_glyphs - 1)
+		next_x_offset = glyphs->glyphs[i+1].geometry.x_offset;
+	      else
+		next_x_offset = 0;
+
+	      dX[num_valid_glyphs] = PANGO_PIXELS (x_offset + next_x_offset) - this_x;
+
+	      /* Prepare for next glyph */
+	      this_x += dX[num_valid_glyphs];
+	      num_valid_glyphs++;
+	    }
+	  i++;
+	}
+#ifdef PANGO_WIN32_DEBUGGING
+      if (pango_win32_debug)
+	{
+	  printf ("ExtTextOutW at %d,%d deltas:",
+		  x + PANGO_PIXELS (start_x_offset),
+		  y + PANGO_PIXELS (cur_y_offset));
+	  for (j = 0; j < num_valid_glyphs; j++)
+	    printf (" %d", dX[j]);
+	  printf ("\n");
+	}
+#endif
+
+      ExtTextOutW (hdc,
+		   x + PANGO_PIXELS (start_x_offset),
+		   y + PANGO_PIXELS (cur_y_offset),
+		   ETO_GLYPH_INDEX,
+		   NULL,
+		   glyph_indexes, num_valid_glyphs,
+		   dX);
+      x += this_x;
     }
 
-  ExtTextOutW (hdc, x, y,
-	       ETO_GLYPH_INDEX,
-	       NULL,
-	       glyph_indexes, glyphs->num_glyphs,
-	       dX);
 
   SelectObject (hdc, old_hfont); /* restore */
   g_free (glyph_indexes);
@@ -278,6 +410,9 @@ pango_win32_font_get_glyph_extents (PangoFont      *font,
   MAT2 m = {{0,1}, {0,0}, {0,0}, {0,1}};
   PangoWin32GlyphInfo *info;
 
+  if (glyph & PANGO_WIN32_UNKNOWN_FLAG)
+    glyph_index = glyph = 0;
+
   info = g_hash_table_lookup (win32font->glyph_info, GUINT_TO_POINTER (glyph));
   
   if (!info)
@@ -309,9 +444,12 @@ pango_win32_font_get_glyph_extents (PangoFont      *font,
   
       if (res == GDI_ERROR)
 	{
-	  guint32 error = GetLastError ();
-	  g_warning ("GetGlyphOutline() failed with error: %d\n", error);
-	  return;
+	  gchar *error = g_win32_error_message (GetLastError ());
+	  g_warning ("GetGlyphOutline(%04X) failed: %s\n",
+		     glyph_index, error);
+	  g_free (error);
+
+	  /* Don't just return now, use the still zeroed out gm */
 	}
 
       info->ink_rect.x = PANGO_SCALE * gm.gmptGlyphOrigin.x;
@@ -334,9 +472,33 @@ pango_win32_font_get_glyph_extents (PangoFont      *font,
     *logical_rect = info->logical_rect;
 }
 
+static int
+max_glyph_width (PangoLayout *layout)
+{
+  int max_width = 0;
+  GSList *l, *r;
+
+  for (l = pango_layout_get_lines (layout); l; l = l->next)
+    {
+      PangoLayoutLine *line = l->data;
+
+      for (r = line->runs; r; r = r->next)
+        {
+          PangoGlyphString *glyphs = ((PangoGlyphItem *)r->data)->glyphs;
+          int i;
+
+          for (i = 0; i < glyphs->num_glyphs; i++)
+            if (glyphs->glyphs[i].geometry.width > max_width)
+              max_width = glyphs->glyphs[i].geometry.width;
+        }
+    }
+
+  return max_width;
+}
+
 static PangoFontMetrics *
-pango_win32_font_get_metrics (PangoFont        *font,
-			      PangoLanguage    *language)
+pango_win32_font_get_metrics (PangoFont     *font,
+			      PangoLanguage *language)
 {
   HFONT hfont;
   TEXTMETRIC tm;
@@ -356,30 +518,31 @@ pango_win32_font_get_metrics (PangoFont        *font,
 
   if (hfont != NULL)
     {
+      PangoFontDescription *font_desc;
+	    
       SelectObject (pango_win32_hdc, hfont);
       GetTextMetrics (pango_win32_hdc, &tm);
 
       metrics->ascent = tm.tmAscent * PANGO_SCALE;
       metrics->descent = tm.tmDescent * PANGO_SCALE;
-      metrics->approximate_digit_width = /* really an approximation ... */
       metrics->approximate_char_width = tm.tmAveCharWidth * PANGO_SCALE;
 
-      /* lovely copy&paste programming (from pangox.c) */
-      /* This is sort of a sledgehammer solution, but we cache this
-       * stuff so not a huge deal hopefully. Get the avg. width of the
-       * chars in "0123456789"
+      /* lovely copy&paste programming (from pangox.c) FIXME: This is
+       * sort of a sledgehammer solution, if we cached the results,
+       * like the other backends did, it wouldn't big a huge performance
+       * problem.  Get the avg. width of the chars in "0123456789"
        */
       context = pango_win32_get_context ();
       pango_context_set_language (context, language);
+      font_desc = pango_font_describe (font);
+      pango_context_set_font_description (context, font_desc);
       layout = pango_layout_new (context);
       pango_layout_set_text (layout, "0123456789", -1);
+      metrics->approximate_digit_width = max_glyph_width (layout);
 
-      pango_layout_get_extents (layout, NULL, &extents);
-   
-      metrics->approximate_digit_width = extents.width / 10.0;
-
-      g_object_unref (G_OBJECT (layout));
-      g_object_unref (G_OBJECT (context));
+      pango_font_description_free (font_desc);
+      g_object_unref (layout);
+      g_object_unref (context);
     }
 
   return metrics;
@@ -422,7 +585,7 @@ pango_win32_font_dispose (GObject *object)
   if (!win32font->in_cache && win32font->fontmap)
     pango_win32_fontmap_cache_add (win32font->fontmap, win32font);
 
-  G_OBJECT_CLASS (parent_class)->dispose (object);
+  G_OBJECT_CLASS (pango_win32_font_parent_class)->dispose (object);
 }
 
 static void
@@ -434,14 +597,14 @@ pango_win32_font_finalize (GObject *object)
   if (win32font->hfont != NULL)
     pango_win32_font_cache_unload (cache, win32font->hfont);
 
-  if (win32font->entry)
-    pango_win32_font_entry_remove (win32font->entry, PANGO_FONT (win32font));
+  if (win32font->win32face)
+    pango_win32_font_entry_remove (win32font->win32face, PANGO_FONT (win32font));
  
   g_hash_table_destroy (win32font->glyph_info);
  
-  g_object_unref (G_OBJECT (win32font->fontmap));
+  g_object_unref (win32font->fontmap);
 
-  G_OBJECT_CLASS (parent_class)->finalize (object);
+  G_OBJECT_CLASS (pango_win32_font_parent_class)->finalize (object);
 }
 
 static PangoFontDescription *
@@ -450,7 +613,7 @@ pango_win32_font_describe (PangoFont *font)
   PangoFontDescription *desc;
   PangoWin32Font *win32font = PANGO_WIN32_FONT (font);
 
-  desc = pango_font_description_copy (win32font->entry->description);
+  desc = pango_font_description_copy (win32font->win32face->description);
   pango_font_description_set_size (desc, win32font->size);
   
   return desc;
@@ -472,28 +635,28 @@ pango_win32_get_shaper_map (PangoLanguage *lang)
 }
 
 static PangoCoverage *
-pango_win32_font_get_coverage (PangoFont  *font,
+pango_win32_font_get_coverage (PangoFont     *font,
 			       PangoLanguage *lang)
 {
   PangoCoverage *coverage;
   PangoWin32Font *win32font = (PangoWin32Font *)font;
 
-  coverage = pango_win32_font_entry_get_coverage (win32font->entry);
+  coverage = pango_win32_font_entry_get_coverage (win32font->win32face, lang);
   if (!coverage)
     {
       coverage = pango_coverage_new ();
-      pango_win32_font_calc_coverage (font, coverage);
+      pango_win32_font_calc_coverage (font, coverage, lang);
       
-      pango_win32_font_entry_set_coverage (win32font->entry, coverage);
+      pango_win32_font_entry_set_coverage (win32font->win32face, coverage, lang);
     }
 
   return coverage;
 }
 
 static PangoEngineShape *
-pango_win32_font_find_shaper (PangoFont   *font,
+pango_win32_font_find_shaper (PangoFont     *font,
 			      PangoLanguage *lang,
-			      guint32      ch)
+			      guint32        ch)
 {
   PangoMap *shape_map = NULL;
 
@@ -506,15 +669,18 @@ pango_win32_font_find_shaper (PangoFont   *font,
 /**
  * pango_win32_get_unknown_glyph:
  * @font: a #PangoFont
+ * @wc: the Unicode character for which a glyph is needed.
  * 
- * Return the index of a glyph suitable for drawing unknown characters.
+ * Returns the index of a glyph suitable for drawing @wc as an
+ * unknown character.
  * 
  * Return value: a glyph index into @font
  **/
 PangoGlyph
-pango_win32_get_unknown_glyph (PangoFont *font)
+pango_win32_get_unknown_glyph (PangoFont *font,
+			       gunichar   wc)
 {
-  return 0;
+  return wc | PANGO_WIN32_UNKNOWN_FLAG;
 }
 
 /**
@@ -527,10 +693,10 @@ pango_win32_get_unknown_glyph (PangoFont *font)
  * Render a #PangoLayoutLine onto a device context
  */
 void 
-pango_win32_render_layout_line (HDC               hdc,
-				PangoLayoutLine  *line,
-				int               x, 
-				int               y)
+pango_win32_render_layout_line (HDC              hdc,
+				PangoLayoutLine *line,
+				int              x, 
+				int              y)
 {
   GSList *tmp_list = line->runs;
   PangoRectangle overall_rect;
@@ -570,10 +736,10 @@ pango_win32_render_layout_line (HDC               hdc,
 					 (bg_color.color.green + 128) >> 8,
 					 (bg_color.color.blue + 128) >> 8));
 	  oldbrush = SelectObject (hdc, brush);
-	  Rectangle (hdc, x + (x_off + logical_rect.x) / PANGO_SCALE,
-			  y + overall_rect.y / PANGO_SCALE,
-			  logical_rect.width / PANGO_SCALE,
-			  overall_rect.height / PANGO_SCALE);
+	  Rectangle (hdc, x + PANGO_PIXELS (x_off + logical_rect.x),
+			  y + PANGO_PIXELS (overall_rect.y),
+			  PANGO_PIXELS (logical_rect.width),
+			  PANGO_PIXELS (overall_rect.height));
 	  SelectObject (hdc, oldbrush);
 	  DeleteObject (brush);
 	}
@@ -587,26 +753,49 @@ pango_win32_render_layout_line (HDC               hdc,
 	}
 
       pango_win32_render (hdc, run->item->analysis.font, run->glyphs,
-			  x + x_off / PANGO_SCALE, y);
+			  x + PANGO_PIXELS (x_off), y);
 
       switch (uline)
 	{
 	case PANGO_UNDERLINE_NONE:
 	  break;
 	case PANGO_UNDERLINE_DOUBLE:
-	  points[0].x = x + (x_off + ink_rect.x) / PANGO_SCALE - 1;
+	  points[0].x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
 	  points[0].y = points[1].y = y + 4;
-	  points[1].x = x + (x_off + ink_rect.x + ink_rect.width) / PANGO_SCALE;
+	  points[1].x = x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width);
 	  Polyline (hdc, points, 2);
 	  /* Fall through */
 	case PANGO_UNDERLINE_SINGLE:
 	  points[0].y = points[1].y = y + 2;
 	  Polyline (hdc, points, 2);
 	  break;
+	case PANGO_UNDERLINE_ERROR:
+          {
+            int point_x;
+            int counter = 0;
+	    int end_x = x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width);
+
+            for (point_x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
+                 point_x <= end_x;
+                 point_x += 2)
+            {
+	      points[0].x = point_x;
+	      points[1].x = MAX (point_x + 1, end_x);
+
+              if (counter)
+	        points[0].y = points[1].y = y + 2;
+              else
+	        points[0].y = points[1].y = y + 3;
+
+	      Polyline (hdc, points, 2);
+              counter = (counter + 1) % 2;
+            }
+          }
+	  break;
 	case PANGO_UNDERLINE_LOW:
-	  points[0].x = x + (x_off + ink_rect.x) / PANGO_SCALE - 1;
-	  points[0].y = points[1].y = y + (ink_rect.y + ink_rect.height) / PANGO_SCALE + 2;
-	  points[1].x = x + (x_off + ink_rect.x + ink_rect.width) / PANGO_SCALE;
+	  points[0].x = x + PANGO_PIXELS (x_off + ink_rect.x) - 1;
+	  points[0].y = points[1].y = y + PANGO_PIXELS (ink_rect.y + ink_rect.height) + 2;
+	  points[1].x = x + PANGO_PIXELS (x_off + ink_rect.x + ink_rect.width);
 	  Polyline (hdc, points, 2);
 	  break;
 	}
@@ -631,77 +820,37 @@ pango_win32_render_layout_line (HDC               hdc,
  * Render a #PangoLayoutLine onto an X drawable
  */
 void 
-pango_win32_render_layout (HDC              hdc,
-			   PangoLayout     *layout,
-			   int              x, 
-			   int              y)
+pango_win32_render_layout (HDC          hdc,
+			   PangoLayout *layout,
+			   int          x, 
+			   int          y)
 {
-  PangoRectangle logical_rect;
-  GSList *tmp_list;
-  PangoAlignment align;
-  int indent;
-  int width;
-  int y_offset = 0;
+  PangoLayoutIter *iter;
 
-  gboolean first = TRUE;
-  
-  g_return_if_fail (layout != NULL);
+  g_return_if_fail (hdc != NULL);
+  g_return_if_fail (PANGO_IS_LAYOUT (layout));
 
-  indent = pango_layout_get_indent (layout);
-  width = pango_layout_get_width (layout);
-  align = pango_layout_get_alignment (layout);
+  iter = pango_layout_get_iter (layout);
 
-  if (width == -1 && align != PANGO_ALIGN_LEFT)
+  do
     {
-      pango_layout_get_extents (layout, NULL, &logical_rect);
-      width = logical_rect.width;
-    }
-  
-  tmp_list = pango_layout_get_lines (layout);
-  while (tmp_list)
-    {
-      PangoLayoutLine *line = tmp_list->data;
-      int x_offset;
+      PangoRectangle   logical_rect;
+      PangoLayoutLine *line;
+      int              baseline;
       
-      pango_layout_line_get_extents (line, NULL, &logical_rect);
-
-      if (width != 1 && align == PANGO_ALIGN_RIGHT)
-	x_offset = width - logical_rect.width;
-      else if (width != 1 && align == PANGO_ALIGN_CENTER)
-	x_offset = (width - logical_rect.width) / 2;
-      else
-	x_offset = 0;
-
-      if (first)
-	{
-	  if (indent > 0)
-	    {
-	      if (align == PANGO_ALIGN_LEFT)
-		x_offset += indent;
-	      else
-		x_offset -= indent;
-	    }
-
-	  first = FALSE;
-	}
-      else
-	{
-	  if (indent < 0)
-	    {
-	      if (align == PANGO_ALIGN_LEFT)
-		x_offset -= indent;
-	      else
-		x_offset += indent;
-	    }
-	}
-	  
-      pango_win32_render_layout_line (hdc, line,
-				      x + x_offset / PANGO_SCALE,
-				      y + (y_offset - logical_rect.y) / PANGO_SCALE);
-
-      y_offset += logical_rect.height;
-      tmp_list = tmp_list->next;
+      line = pango_layout_iter_get_line (iter);
+      
+      pango_layout_iter_get_line_extents (iter, NULL, &logical_rect);
+      baseline = pango_layout_iter_get_baseline (iter);
+      
+      pango_win32_render_layout_line (hdc,
+                                      line,
+                                      x + PANGO_PIXELS (logical_rect.x),
+                                      y + PANGO_PIXELS (baseline));
     }
+  while (pango_layout_iter_next_line (iter));
+
+  pango_layout_iter_free (iter);  
 }
 
 /* This utility function is duplicated here and in pango-layout.c; should it be
@@ -757,48 +906,11 @@ pango_win32_get_item_properties (PangoItem      *item,
     }
 }
 
-
-
-/* Various truetype defines: */
-
-#define MAKE_TT_TABLE_NAME(c1, c2, c3, c4) \
-   (((guint32)c4) << 24 | ((guint32)c3) << 16 | ((guint32)c2) << 8 | ((guint32)c1))
-
-#define CMAP (MAKE_TT_TABLE_NAME('c','m','a','p'))
-#define CMAP_HEADER_SIZE 4
-#define ENCODING_TABLE_SIZE 8
-#define TABLE_4_HEADER_SIZE (sizeof (guint16)*6)
-
-#define UNICODE_PLATFORM_ID 3
-#define UNICODE_SPECIFIC_ID 1
-
-#pragma pack(1)
-
-struct encoding_table { /* Must be packed! */
-  guint16 platform_id;
-  guint16 encoding_id;
-  guint32 offset;
-};
-
-struct type_4_table { /* Must be packed! */
-  guint16 format;
-  guint16 length;
-  guint16 version;
-  guint16 seg_count_x_2;
-  guint16 search_range;
-  guint16 entry_selector;
-  guint16 range_shift;
-
-  guint16 reserved;
-
-  guint16 arrays[1];
-};
-
 static guint 
 get_unicode_mapping_offset (HDC hdc)
 {
   guint16 n_tables;
-  struct encoding_table *table;
+  struct cmap_encoding_subtable *table;
   gint32 res;
   int i;
   guint32 offset;
@@ -818,8 +930,8 @@ get_unicode_mapping_offset (HDC hdc)
 
   for (i = 0; i < n_tables; i++)
     {
-      if (table[i].platform_id == GUINT16_TO_BE (UNICODE_PLATFORM_ID) &&
-	  table[i].encoding_id == GUINT16_TO_BE (UNICODE_SPECIFIC_ID))
+      if (table[i].platform_id == GUINT16_TO_BE (MICROSOFT_PLATFORM_ID) &&
+	  table[i].encoding_id == GUINT16_TO_BE (UNICODE_ENCODING_ID))
 	{
 	  offset = GUINT32_FROM_BE (table[i].offset);
 	  g_free (table);
@@ -830,14 +942,14 @@ get_unicode_mapping_offset (HDC hdc)
   return 0;
 }
 
-static struct type_4_table *
+static struct type_4_cmap *
 get_unicode_mapping (HDC hdc)
 {
   guint32 offset;
   guint32 res;
   guint16 length;
   guint16 *tbl, *tbl_end;
-  struct type_4_table *table;
+  struct type_4_cmap *table;
 
   /* FIXME: Could look here at the CRC for the font in the DC 
             and return a cached copy if the same */
@@ -862,7 +974,7 @@ get_unicode_mapping (HDC hdc)
   
   table->format = GUINT16_FROM_BE (table->format);
   table->length = GUINT16_FROM_BE (table->length);
-  table->version = GUINT16_FROM_BE (table->version);
+  table->language = GUINT16_FROM_BE (table->language);
   table->seg_count_x_2 = GUINT16_FROM_BE (table->seg_count_x_2);
   table->search_range = GUINT16_FROM_BE (table->search_range);
   table->entry_selector = GUINT16_FROM_BE (table->entry_selector);
@@ -887,28 +999,28 @@ get_unicode_mapping (HDC hdc)
 }
 
 static guint16 *
-get_id_range_offset (struct type_4_table *table)
+get_id_range_offset (struct type_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   return &table->arrays[seg_count*3];
 }
 
 static guint16 *
-get_id_delta (struct type_4_table *table)
+get_id_delta (struct type_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   return &table->arrays[seg_count*2];
 }
 
 static guint16 *
-get_start_count (struct type_4_table *table)
+get_start_count (struct type_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   return &table->arrays[seg_count*1];
 }
 
 static guint16 *
-get_end_count (struct type_4_table *table)
+get_end_count (struct type_4_cmap *table)
 {
   gint32 seg_count = table->seg_count_x_2/2;
   /* Apparently the reseved spot is not reserved for 
@@ -918,9 +1030,9 @@ get_end_count (struct type_4_table *table)
 
 
 static gboolean
-find_segment (struct type_4_table *table, 
-	      guint16              wc,
-	      guint16             *segment)
+find_segment (struct type_4_cmap *table, 
+	      guint16             wc,
+	      guint16            *segment)
 {
   guint16 start, end, i;
   guint16 seg_count = table->seg_count_x_2/2;
@@ -978,21 +1090,21 @@ find_segment (struct type_4_table *table,
   return FALSE;
 }
 
-static struct type_4_table *
+static struct type_4_cmap *
 font_get_unicode_table (PangoFont *font)
 {
   PangoWin32Font *win32font = (PangoWin32Font *)font;
   HFONT hfont;
-  struct type_4_table *table;
+  struct type_4_cmap *table;
 
-  if (win32font->entry->unicode_table)
-    return (struct type_4_table *)win32font->entry->unicode_table;
+  if (win32font->win32face->unicode_table)
+    return (struct type_4_cmap *)win32font->win32face->unicode_table;
 
   hfont = pango_win32_get_hfont (font);
   SelectObject (pango_win32_hdc, hfont);
   table = get_unicode_mapping (pango_win32_hdc);
   
-  win32font->entry->unicode_table = table;
+  win32font->win32face->unicode_table = table;
 
   return table;
 }
@@ -1010,7 +1122,7 @@ gint
 pango_win32_font_get_glyph_index (PangoFont *font,
 				  gunichar   wc)
 {
-  struct type_4_table *table;
+  struct type_4_cmap *table;
   guint16 *id_range_offset;
   guint16 *id_delta;
   guint16 *start_count;
@@ -1047,11 +1159,106 @@ pango_win32_font_get_glyph_index (PangoFont *font,
   return glyph;
 }
 
+gboolean
+pango_win32_get_name_header (HDC                 hdc,
+			     struct name_header *header)
+{
+  if (GetFontData (hdc, NAME, 0, header, sizeof (*header)) != sizeof (*header))
+    return FALSE;
+
+  header->num_records = GUINT16_FROM_BE (header->num_records);
+  header->string_storage_offset = GUINT16_FROM_BE (header->string_storage_offset);
+
+  return TRUE;
+}
+  
+gboolean
+pango_win32_get_name_record (HDC                 hdc,
+			     gint                i,
+			     struct name_record *record)
+{
+  if (GetFontData (hdc, NAME, 6 + i * sizeof (*record),
+		   record, sizeof (*record)) != sizeof (*record))
+    return FALSE;
+  
+  record->platform_id = GUINT16_FROM_BE (record->platform_id);
+  record->encoding_id = GUINT16_FROM_BE (record->encoding_id);
+  record->language_id = GUINT16_FROM_BE (record->language_id);
+  record->name_id = GUINT16_FROM_BE (record->name_id);
+  record->string_length = GUINT16_FROM_BE (record->string_length);
+  record->string_offset = GUINT16_FROM_BE (record->string_offset);
+
+  return TRUE;
+}
+
+static gboolean
+font_has_name_in (PangoFont                       *font,
+		  PangoWin32CoverageLanguageClass  cjkv)
+{
+  HFONT hfont, oldhfont;
+  struct name_header header;
+  struct name_record record;
+  gint i;
+  gboolean retval = FALSE;
+
+  if (cjkv == PANGO_WIN32_COVERAGE_UNSPEC)
+    return TRUE;
+
+  hfont = pango_win32_get_hfont (font);
+  oldhfont = SelectObject (pango_win32_hdc, hfont);
+
+  if (!pango_win32_get_name_header (pango_win32_hdc, &header))
+    {
+      SelectObject (pango_win32_hdc, oldhfont);
+      return FALSE;
+    }
+
+  for (i = 0; i < header.num_records; i++)
+    {
+      if (!pango_win32_get_name_record (pango_win32_hdc, i, &record))
+	{
+	  SelectObject (pango_win32_hdc, oldhfont);
+	  return FALSE;
+	}
+
+      if ((record.name_id != 1 && record.name_id != 16) || record.string_length <= 0)
+	continue;
+
+      PING(("platform:%d encoding:%d language:%04x name_id:%d",
+	    record.platform_id, record.encoding_id, record.language_id, record.name_id));
+
+      if (record.platform_id == MICROSOFT_PLATFORM_ID)
+	if ((cjkv == PANGO_WIN32_COVERAGE_ZH_TW &&
+	     record.language_id == MAKELANGID (LANG_CHINESE, SUBLANG_CHINESE_TRADITIONAL))
+	    ||
+	    (cjkv == PANGO_WIN32_COVERAGE_ZH_CN &&
+	     record.language_id == MAKELANGID (LANG_CHINESE, SUBLANG_CHINESE_SIMPLIFIED))
+	    ||
+	    (cjkv == PANGO_WIN32_COVERAGE_JA &&
+	     PRIMARYLANGID (record.language_id) == LANG_JAPANESE)
+	    ||
+	    (cjkv == PANGO_WIN32_COVERAGE_KO &&
+	     PRIMARYLANGID (record.language_id) == LANG_KOREAN)
+	    ||
+	    (cjkv == PANGO_WIN32_COVERAGE_VI &&
+	     PRIMARYLANGID (record.language_id) == LANG_VIETNAMESE))
+	  {
+	    PING (("yep:%d:%04x", cjkv, record.language_id));
+	    retval = TRUE;
+	    break;
+	  }
+    }
+
+  SelectObject (pango_win32_hdc, oldhfont);
+  return retval;
+}
+
 static void
 pango_win32_font_calc_coverage (PangoFont     *font,
-				PangoCoverage *coverage)
+				PangoCoverage *coverage,
+				PangoLanguage *lang)
 {
-  struct type_4_table *table;
+  struct type_4_cmap *table;
   guint16 *id_range_offset;
   guint16 *start_count;
   guint16 *end_count;
@@ -1059,42 +1266,114 @@ pango_win32_font_calc_coverage (PangoFont     *font,
   guint16 id;
   guint32 ch;
   int i;
+  PangoWin32CoverageLanguageClass cjkv;
+  gboolean hide_unihan = FALSE;
+  
+  PING(("font:%s lang:%s",
+	pango_font_description_to_string (pango_font_describe (font)),
+	pango_language_to_string (lang)));
+
+  cjkv = pango_win32_coverage_language_classify (lang);
+
+  if (cjkv != PANGO_WIN32_COVERAGE_UNSPEC && !font_has_name_in (font, cjkv))
+    {
+      PING(("hiding UniHan chars"));
+      hide_unihan = TRUE;
+    }
 
   /* Do GetFontData magic on font->hfont here. */
 
   table = font_get_unicode_table (font);
 
   if (table == NULL)
-    return;
+    {
+      PING(("no table"));
+      return;
+    }
   
   seg_count = table->seg_count_x_2/2;
   end_count = get_end_count (table);
   start_count = get_start_count (table);
   id_range_offset = get_id_range_offset (table);
 
+  PING (("coverage:"));
   for (i = 0; i < seg_count; i++)
     {
       if (id_range_offset[i] == 0)
 	{
+#ifdef PANGO_WIN32_DEBUGGING
+	  if (pango_win32_debug)
+	    {
+	      if (end_count[i] == start_count[i])
+		printf ("%04x ", start_count[i]);
+	      else
+		printf ("%04x:%04x ", start_count[i], end_count[i]);
+	    }
+#endif	  
 	  for (ch = start_count[i]; 
 	       ch <= end_count[i];
 	       ch++)
-	    pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
+	    if (hide_unihan && ch >= 0x3400 && ch <= 0x9FAF)
+	      pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
+	    else
+	      pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
 	}
       else
 	{
+#ifdef PANGO_WIN32_DEBUGGING
+	  guint32 ch0 = G_MAXUINT;
+#endif
 	  for (ch = start_count[i]; 
 	       ch <= end_count[i];
 	       ch++)
 	    {
+	      if (ch == 0xFFFF)
+		break;
+
 	      id = *(id_range_offset[i]/2 +
 		     (ch - start_count[i]) +
 		     &id_range_offset[i]);
 	      if (id)
-		pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
+		{
+#ifdef PANGO_WIN32_DEBUGGING
+		  if (ch0 == G_MAXUINT)
+		    ch0 = ch;
+#endif
+		  if (hide_unihan && ch >= 0x3400 && ch <= 0x9FAF)
+		    pango_coverage_set (coverage, ch, PANGO_COVERAGE_APPROXIMATE);
+		  else
+		    pango_coverage_set (coverage, ch, PANGO_COVERAGE_EXACT);
+		}
+#ifdef PANGO_WIN32_DEBUGGING
+	      else if (ch0 < G_MAXUINT)
+		{
+		  if (pango_win32_debug)
+		    {
+		      if (ch > ch0 + 2)
+			printf ("%04x:%04x ", ch0, ch - 1);
+		      else
+			printf ("%04x ", ch0);
+		    }
+		  ch0 = G_MAXUINT;
+		}
+#endif
 	    }
+#ifdef PANGO_WIN32_DEBUGGING
+	  if (ch0 < G_MAXUINT)
+	    {
+	      if (pango_win32_debug)
+		{
+		  if (ch > ch0 + 2)
+		    printf ("%04x:%04x ", ch0, ch - 1);
+		  else
+		    printf ("%04x ", ch0);
+		}
+	    }
+#endif
 	}
     }
+#ifdef PANGO_WIN32_DEBUGGING
+  if (pango_win32_debug)
+    printf ("\n");
+#endif
 }
-
-
